package be.cylab.java.wowa.training;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;

/**
 * Class Trainer. Learn wowa weights based on training dataset
 */
public class Trainer {

    private final TrainerParameters parameters;
    private final Factory factory;

    /**
     * Trainer constructor.
     *
     * @param parameters TrainerParameters Parameters for trainer
     */
    public Trainer(final TrainerParameters parameters,
                   final AbstractSolution solution_type) {
        this.parameters = parameters;
        this.factory = new Factory(solution_type);
    }

    /**
     * @param data
     * @param expected
     * @return
     */
    public final AbstractSolution run(
            final List<double[]> data,
            final double[] expected) {

        List<AbstractSolution> current_population
                = this.generateInitialPopulationAndComputeDistances(
                this.getParameters().getPopulationSize(),
                data,
                expected);

        //Collections.sort(current_population);
        AbstractSolution best_solution
                = this.findBestSolution(current_population);
        for (int generation = 0;
             generation < this.parameters.getMaxGenerationNumber();
             generation++) {
            current_population = this.performReproduction(
                    current_population, data, expected);
            AbstractSolution best_solution_of_current_population
                    = this.findBestSolution(current_population);

            if (this.getParameters().getLogger() != null) {
                this.getParameters().getLogger().log(
                        Level.INFO,
                        "Generation " + (generation + 1) + " : "
                                + best_solution.toString());
            }

            //We found a new best solution
            if (best_solution_of_current_population.getDistance()
                    < best_solution.getDistance()) {
                best_solution = best_solution_of_current_population;
            }

        }
        return best_solution;
    }

    /**
     * @param solutions
     * @return
     */
    final AbstractSolution findBestSolution(
            final List<AbstractSolution> solutions) {

        AbstractSolution best_solution = solutions.get(0);

        for (AbstractSolution solution : solutions) {
            if (solution.getDistance() < best_solution.getDistance()) {
                best_solution = solution;
            }
        }
        return best_solution;
    }

    /**
     * @param number_of_weights
     * @param population_size
     * @return
     */
    final List<AbstractSolution> generateInitialPopulation(
            final int number_of_weights,
            final int population_size) {

        List<AbstractSolution> population = new ArrayList<>();
        for (int i = 0; i < population_size; i++) {
            AbstractSolution solution = this.factory.createSolutionObject(number_of_weights);
            population.add(solution);
        }
        return population;
    }

    /**
     * This method generates an initial population with specific element.
     * A part of the population is random (as generateInitialPopulation)
     * The last elements are specific elements like [0, 0, 1, 0], [1, 0, 0, 0].
     * All weights in a vector are null except one that is equal to 1.
     * There are number_of_weight^2 specific elements or half of the population
     * Depends on the population size and the number of weight
     * @param number_of_weight
     * @param population_size
     * @return
     */
    final List<AbstractSolution> generateQuasiRandomInitialPopulation(
            final int number_of_weight,
            final int population_size) {
        int number_of_random_element = population_size / 2;
        if (Math.pow(number_of_weight, 2) < population_size / 2) {
            number_of_random_element =
                    (int) (population_size - Math.pow(number_of_weight, 2));
        }
        //Random part of the population
        List<AbstractSolution> population = this.generateInitialPopulation(
                number_of_weight,
                number_of_random_element);
        int[][] random_memory = new int[number_of_weight][number_of_weight];
        while (population.size() < population_size) {
            int position_w = Utils.randomInteger(0, number_of_weight - 1);
            int position_p = Utils.randomInteger(0, number_of_weight - 1);
            if (random_memory[position_w][position_p] == 0) {
                AbstractSolution solution = new SolutionDistance(5);
                double[] weights_w = new double[number_of_weight];
                weights_w[position_w] = 1;
                double[] weights_p = new double[number_of_weight];
                solution.setWeightsW(weights_w);
                solution.setWeightsP(weights_p);
                population.add(solution);
                random_memory[position_w][position_p] = 1;
            }
        }
        return population;

    }

    /**
     * @param solutions
     * @param data
     * @param expected
     * @return
     */
    final List<AbstractSolution> computeDistances(
            final List<AbstractSolution> solutions,
            final List<double[]> data,
            final double[] expected) {

        for (AbstractSolution solution : solutions) {
            solution.computeScoreTo(data, expected);
        }
        return solutions;

    }

    /**
     * Select elements for the next generation.
     * Each element has a probability to be chosen proportional to its
     * fitness score (distance for example)
     * @param solutions
     * @param selected_elements
     * @param count
     * @return
     */
    final List<AbstractSolution> rouletteWheelSelection(
            final List<AbstractSolution> solutions,
            final List<AbstractSolution> selected_elements,
            final int count) {

        double max = Utils.findMaxDistance(solutions);
        double min = Utils.findMinDistance(solutions);
        double sum = Utils.sumTotalDistance(solutions);

        if (solutions.size() < count) {
            throw new IllegalArgumentException(
                    "Not enough elements in population to select "
                            + count + "parents"
            );
        }

        while (selected_elements.size() < count) {
            double tos = Math.random();
            double normalized_distance = 0;
            for (AbstractSolution solution : solutions) {
                normalized_distance = normalized_distance
                        + (max + min - solution.getDistance()) / sum;
                if (normalized_distance > tos) {
                    selected_elements.add(solution);
                    solutions.remove(solution);
                    break;
                }
            }
        }
        return selected_elements;
    }

    /**
     * Select elements for the next generation.
     * Two elements are randomly selected. The element with the best
     * fitness score is selected. The other is put again in the population
     * @param solutions
     * @param selected_elements
     * @param count
     * @return
     */
    final List<AbstractSolution> tournamentSelection(
            final List<AbstractSolution> solutions,
            final List<AbstractSolution> selected_elements,
            final int count) {

        while (selected_elements.size() < count) {
            //Collections.sort(solutions);

            //Select random element in the population
            int solution1_index = Utils.randomInteger(0, solutions.size() - 1);
            AbstractSolution solution1 = solutions.get(solution1_index);

            //Select random element in the population
            int solution2_index = Utils.randomInteger(0, solutions.size() - 1);
            AbstractSolution solution2 = solutions.get(solution2_index);

            //Compare the two selected element and put the solution with
            // the smallest distance in selected list
            if (solution1.getDistance() < solution2.getDistance()) {
                selected_elements.add(solution1);
                solutions.remove(solution1);
            } else {
                selected_elements.add(solution2);
                solutions.remove(solution2);

            }
        }
        return selected_elements;
    }

    /**
     * Method to select parents for the mating.
     * Use an elitism mechanism. The two best elements of the current population
     * are systematically selected for the next one.
     * @param solutions
     * @param count
     * @param selection_method
     * @return
     */
    final List<AbstractSolution> selectParents(
            final List<AbstractSolution> solutions,
            final int count,
            final int selection_method) {

        List<AbstractSolution> selected_parents = new ArrayList<>();
        //Select the two best current solutions

        //Sort population
        Collections.sort(solutions);
        //Put the two best elements in selected parents list
        selected_parents.add(solutions.get(0));
        selected_parents.add(solutions.get(1));
        //Remove the two best elements from the list solutions
        solutions.remove(0);
        solutions.remove(1);

        if (selection_method == TrainerParameters.SELECTION_METHOD_RWS) {

            return this.rouletteWheelSelection(solutions,
                    selected_parents,
                    count
            );
        } else if (selection_method == TrainerParameters.SELECTION_METHOD_TOS) {
            return this.tournamentSelection(solutions,
                    selected_parents,
                    count
            );
        }
        throw new IllegalArgumentException("Invalid selection method");

    }

    /**
     * Select two random elements in the selected parents population.
     * Reproduce these two parents to generate 2 news individuals
     * @param solutions
     * @return
     */
    final List<AbstractSolution> doReproduction(
            final List<AbstractSolution> solutions) {

        int original_solution_size = solutions.size();
        int nbr_weights = solutions.get(0).getWeightsP().length;

        //we add children to the current list of solutions
        while (solutions.size() < this.parameters.getPopulationSize()) {
            int dad = Utils.randomInteger(0, original_solution_size - 1);
            int mom;
            do {
                mom = Utils.randomInteger(0, original_solution_size - 1);
            } while (dad == mom);

            int cut_position = Utils.randomInteger(0, nbr_weights - 1);

            //beta is used to compute the new value at the cut position
            double beta = Math.random();
            this.reproduce(solutions.get(dad),
                    solutions.get(mom),
                    solutions,
                    cut_position,
                    beta
            );
        }
        while (solutions.size() > this.getParameters().getPopulationSize()) {
            solutions.remove(solutions.get(solutions.size() - 1));
        }
        return solutions;
    }

    /**
     * Two selected parents are combined to generate two new elements.
     * The method used is detailed in
     * https://gitlab.cylab.be/cylab/wowa-training/
     * blob/master/doc/TheContinuousGeneticAlgorithm.pdf p59
     * @param dad
     * @param mom
     * @param solutions
     * @param cut_position
     * @param beta
     * @return
     */
    final void reproduce(final AbstractSolution dad,
                         final AbstractSolution mom,
                         final List<AbstractSolution> solutions,
                         final int cut_position,
                         final double beta) {

        double p_new1_w = dad.getWeightsW()[cut_position] - beta
                * (dad.getWeightsW()[cut_position]
                - mom.getWeightsW()[cut_position]);
        double p_new2_w = mom.getWeightsW()[cut_position] + beta
                * (dad.getWeightsW()[cut_position]
                - mom.getWeightsW()[cut_position]);

        double p_new1_p = dad.getWeightsP()[cut_position] - beta
                * (dad.getWeightsP()[cut_position]
                - mom.getWeightsP()[cut_position]);
        double p_new2_p = mom.getWeightsP()[cut_position] + beta
                * (dad.getWeightsP()[cut_position]
                - mom.getWeightsP()[cut_position]);

        AbstractSolution child1
                = this.factory.createSolutionObject(dad.weights_p.length);
        AbstractSolution child2
                = this.factory.createSolutionObject(dad.weights_p.length);

        for (int i = 0; i < cut_position; i++) {
            child1.getWeightsW()[i] = dad.getWeightsW()[i];
            child1.getWeightsP()[i] = dad.getWeightsP()[i];

            child2.getWeightsW()[i] = mom.getWeightsW()[i];
            child2.getWeightsP()[i] = mom.getWeightsP()[i];
        }

        child1.getWeightsW()[cut_position] = p_new1_w;
        child2.getWeightsW()[cut_position] = p_new2_w;
        child1.getWeightsP()[cut_position] = p_new1_p;
        child2.getWeightsP()[cut_position] = p_new2_p;

        int nbre_weights = dad.getWeightsW().length;
        for (int i = cut_position + 1; i < nbre_weights; i++) {
            child1.getWeightsW()[i] = mom.getWeightsW()[i];
            child1.getWeightsP()[i] = mom.getWeightsP()[i];

            child2.getWeightsW()[i] = dad.getWeightsW()[i];
            child2.getWeightsP()[i] = dad.getWeightsP()[i];
        }
        //Check and correct only if we used a QuasiRandom generation
        if (this.getParameters().getPopulationInitializationMethod()
                == TrainerParameters.POPULATION_INITIALIZATION_QUASI_RANDOM) {
            this.checkAndCorrectNullWeightVector(child1);
            this.checkAndCorrectNullWeightVector(child2);
        }

        child1.normalize();
        child2.normalize();
        solutions.add(child1);
        solutions.add(child2);
    }

    /**
     * Mutate random individuals with mutation rate probability.
     * @param solutions
     * @return
     */
    final List<AbstractSolution> randomlyMutateGenes(
            final List<AbstractSolution> solutions) {

        double probability =
                (double) this.getParameters().getMutationRate() / 100;

        for (int i = 0; i < solutions.size(); i++) {
            if (i > 1) {
                solutions.get(i).randomlyMutateWithProbability(probability);
            }
        }

        return solutions;
    }

    /**
     * @return
     */
    final TrainerParameters getParameters() {
        return this.parameters;
    }

    /**
     * @param population_size
     * @param data
     * @param expected
     * @return
     */
    private List<AbstractSolution> generateInitialPopulationAndComputeDistances(
            final int population_size,
            final List<double[]> data,
            final double[] expected) {

        int number_of_weights = data.get(0).length;
        List<AbstractSolution> initial_population = new ArrayList<>();
        if (this.parameters.getPopulationInitializationMethod()
                == TrainerParameters.POPULATION_INITIALIZATION_RANDOM) {
            initial_population = this.generateInitialPopulation(
                    number_of_weights,
                    population_size
            );
        } else if (this.parameters.getPopulationInitializationMethod()
                == TrainerParameters.POPULATION_INITIALIZATION_QUASI_RANDOM) {
            initial_population = this.generateQuasiRandomInitialPopulation(
                    number_of_weights,
                    population_size);
        }
        return this.computeDistances(initial_population, data, expected);
    }

    /**
     * @param population
     * @param data
     * @param expected
     * @return
     */
    final List<AbstractSolution> performReproduction(
            final List<AbstractSolution> population,
            final List<double[]> data,
            final double[] expected) {

        List<AbstractSolution> parents = this.selectParents(population,
                this.getParameters().getNumberOfParents(),
                this.getParameters().getSelectionMethod());
        List<AbstractSolution> new_generation = this.doReproduction(parents);
        List<AbstractSolution> mutated
                = this.randomlyMutateGenes(new_generation);

        return this.computeDistances(mutated, data, expected);
    }

    /**
     * Correct weights vectors if it is null.
     * With QUASI_RANDOM initialization, it is possible to obtain a full null
     * weight vector after mating. The method check and correct that
     * @param child AbstractSolution
     */
    private void checkAndCorrectNullWeightVector(final AbstractSolution child) {
        if (Utils.sumArrayElements(child.getWeightsW()) == 0) {
            int index = Utils.randomInteger(0, child.getWeightsW().length) - 1;
            double[] weights_w = new double[5];
            weights_w[index] = 1;
            child.setWeightsW(weights_w);
        }
        if (Utils.sumArrayElements(child.getWeightsP()) == 0) {
            int index = Utils.randomInteger(0, child.getWeightsP().length) - 1;
            double[] weights_p = new double[5];
            weights_p[index] = 1;
            child.setWeightsP(weights_p);
        }
    }
}
